#############################################################################
#
#	core.py - Pyro Core Library
#
#	This is part of "Pyro" - Python Remote Objects
#	which is (c) Irmen de Jong - irmen@users.sourceforge.net
#
#############################################################################

import sys, time, sre, traceback
import Pyro
import util, protocol, constants
from errors import *
from types import UnboundMethodType, MethodType, BuiltinMethodType, TupleType, StringType, UnicodeType

Log=util.Log


#### Has init been performed already?
_init_server_done=0
_init_client_done=0



#############################################################################
#
#	ObjBase		- Server-side object implementation base class
#	              or master class with the actual object as delegate
#
#############################################################################

class ObjBase:
	def __init__(self):
		self.objectGUID=util.getBinaryGUID()
		self.delegate=None
		if Pyro.config.PYRO_MOBILE_CODE:
			self.codeValidator=lambda n,m,a: 1  # always accept
	def GUID(self):
		return self.objectGUID
	def setGUID(self, guid):			# used with persistent name server
		self.objectGUID = guid
	def delegateTo(self,delegate):
		self.delegate=delegate
	def setDaemon(self, daemon):
		self.daemon=daemon
	def setCodeValidator(self, v):
		self.codeValidator=v
	def getDaemon(self):
		return self.daemon
	def getProxy(self):
		return self.daemon.getProxyForObj(self)
	def getAttrProxy(self):
		return self.daemon.getAttrProxyForObj(self)
	def Pyro_dyncall(self, method, flags, args):
		# find the method in this object, and call it with the supplied args.
		keywords={}
		if flags & constants.RIF_Keywords:
			# reconstruct the varargs from a tuple like
			#  (a,b,(va1,va2,va3...),{kw1:?,...})
			keywords=args[-1]
			args=args[:-1]
		if flags & constants.RIF_Varargs:
			# reconstruct the varargs from a tuple like (a,b,(va1,va2,va3...))
			args=args[:-1]+args[-1]
		# If the method is part of ObjBase, never call the delegate object because
		# that object doesn't implement that method. If you don't check this,
		# remote attributes won't work with delegates for instance, because the
		# delegate object doesn't implement _r_xa. (remote_xxxattr)
		if method in dir(ObjBase):  
			return apply(getattr(self,method),args,keywords)
		else:
			return apply(getattr(self.delegate or self,method),args,keywords)
	# remote getattr/setattr support:
	def _r_ha(self, attr):
		try:
			attr = getattr(self.delegate or self,attr)
			if type(attr) in (UnboundMethodType, MethodType, BuiltinMethodType):
				return 1 # method
		except:
			pass
		return 2 # attribute
	def _r_ga(self, attr):
		return getattr(self.delegate or self, attr)
	def _r_sa(self, attr, value):
		setattr(self.delegate or self, attr, value)
	# remote code downloading support:
	def remote_supply_code(self, name, module, sourceaddr):
		if Pyro.config.PYRO_MOBILE_CODE and self.codeValidator(name,module,sourceaddr):
			import imp,marshal,new
			Log.msg('ObjBase','loading supplied code: ',name,'from',str(sourceaddr))
			name=name.split('.')
			# make the module hierarchy and add all names to sys.modules
			path=''
			mod=new.module('agent-context')
			for m in name:
				path+='.'+m
				setattr(mod,m,new.module(path[1:]))
				mod=getattr(mod,m)
				sys.modules[path[1:]]=mod
				
			if module[0:4]!=imp.get_magic():
				# compile source code
				code=compile(module,'<downloaded>','exec')
			else:
				# read bytecode
				code=marshal.loads(module[8:])
			# finally, execute the module code in the right module.	
			exec code in mod.__dict__
		else:
			Log.warn('ObjBase','attempt to supply code denied: ',name,'from',str(sourceaddr))
			raise PyroError('attempt to supply code denied')
			

#############################################################################
#
#	PyroURI		- Pyro Universal Resource Identifier
#
#	This class represents a Pyro URI (which consists of four parts,
#	a protocol identifier, an IP address, a portnumber, and an object ID.
#	
#	The URI can be converted to a string representation (str converter).
#	The URI can also be read back from such a string (reinitFromString).
#	The URI can be initialised from its parts (init).
#	The URI can be initialised from a string directly, if the init
#	 code detects a ':' and '/' in the host argument (which is then
#        assumed to be a string URI, not a host name/ IP address).
#
#############################################################################

class PyroURI:
	def __init__(self,host,objectID=0,port=0,prtcol='PYRO'):
		# if the 'host' arg is a PyroURI, copy contents
		if isinstance(host, PyroURI):
			self.reinitFromString(str(host))
		else:	
			# If the 'host' arg contains ':' and '/', assume it's a URI string.
			if ':' in host and '/' in host:
				self.reinitFromString(host)
			else:
				if '/' in host:
					raise URIError('malformed hostname')
				if Pyro.config.PYRO_DNS_URI:
					self.address = host
				else:
					self.address=protocol.getIPAddress(host)
					if not self.address:
						raise URIError('unknown host')
				if port:
					self.port=port
				else:
					self.port=Pyro.config.PYRO_PORT
				self.protocol=prtcol
				self.objectID=objectID
	def __str__(self):
		if self.protocol=='PYRO':
			return self.protocol+'://'+self.address+':'+str(self.port)+  \
								'/'+util.bin2ascGUID(self.objectID)
		else:						
			return self.protocol+'://'+self.address+':'+str(self.port)+'/'+self.objectID
	def __repr__(self):
		return '<PyroURI \''+str(self)+'\'>'
	def init(self,address,objectID,port=0,prtcol='PYRO'):
		self.address=address
		self.objectID=objectID
		if port:
			self.port=port
		else:
			self.port=Pyro.config.PYRO_PORT
		self.protocol=prtcol
	def reinitFromString(self,arg):
		x=sre.match(r'(?P<protocol>[^\s:/]+)://(?P<hostname>[^\s:]+):?(?P<port>\d+)?/(?P<id>\S*)',arg)
		if x:
			self.protocol=x.group('protocol')
			self.address=x.group('hostname')
			self.port=x.group('port')
			if self.port:
				self.port=int(self.port)
			else:
				self.port=Pyro.config.PYRO_PORT
			if self.protocol=='PYRO':	
				self.objectID=util.asc2binGUID(x.group('id'))
			else:
				self.objectID=x.group('id')
			return
		Log.error('PyroURI','illegal URI format passed: '+arg)
		raise URIError('illegal URI format')



def processStringURI(URI):
	# PYRONAME://[hostname[:port]/]objectname
	x=sre.match(r'PYRONAME://(((?P<hostname>[^\s:]+):(?P<port>\d+)/)|((?P<onlyhostname>[^\s:]+)/))?(?P<name>\S*)',URI)
	if x:
		import naming
		hostname=x.group('hostname') or x.group('onlyhostname') 
		port=x.group('port') 
		name=x.group('name')
		loc=naming.NameServerLocator()
		if port:
			port=int(port)
		NS=loc.getNS(host=hostname,port=port)
		return NS.resolve(name)
	# PYROLOC://hostname[:port]/objectname
	x=sre.match(r'PYROLOC://(?P<hostname>[^\s:]+):?(?P<port>\d+)?/(?P<name>\S*)',URI)
	if x:
		hostname=x.group('hostname')
		port=x.group('port')
		if port:
			port=int(port)
		else:
			port=0
		name=x.group('name')
		return PyroURI(hostname,name,port,'PYROLOC')
	# let the normal Pyro URI deal with it in the regular way
	return PyroURI(URI)


#############################################################################
#
#	DynamicProxy	- dynamic Pyro proxy

#
#	Can be used by clients to invoke objects for which they have no
#	precompiled proxy.
#
#############################################################################

def getProxyForURI(URI):
	return DynamicProxy(URI)
def getAttrProxyForURI(URI):
	return DynamicProxyWithAttrs(URI)

class DynamicProxy:
	def __init__(self, URI):
		if type(URI) in (StringType,UnicodeType):
			URI=processStringURI(URI)
		self.URI = URI
		self.objectID = URI.objectID
		self.adapter = None				# delay adapter binding
		self._name=[]
		self._onewayMethods=[]
	def _setOneway(self, methods):
		if type(methods) not in (type([]), type((0,))):
			methods=(methods,)
		if self.adapter:
			self.adapter.setOneway(methods)
		else:
			self._onewayMethods.extend(methods)
	def __copy__(self):
		c=DynamicProxy(self.URI)
		c.adapter=None
		return c
	def __getattr__(self, name):
		# allows it to be safely pickeled
		if name!="__getinitargs__":
			self._name.append(name)
			return self.__invokePYRO__
		raise AttributeError()
	def __repr__(self):
		return "<"+self.__class__.__name__+" instance at "+hex(id(self))+">"
	def __str__(self):
		return repr(self)
	def __hash__(self):
		# makes it possible to use this class as a key in a dict
		return hash(self.objectID)
	def __lt__(self,other):
		# makes it possible to compare two proxies using objectID
		return self.objectID<other.objectID
	def __gt__(self,other):
		# makes it possible to compare two proxies using objectID
		return self.objectID>other.objectID
	def __eq__(self,other):
		# makes it possible to compare two proxies using objectID
		return self.objectID==other.objectID
	def __cmp__(self,other):
		# makes it possible to compare two proxies using objectID
		return cmp(self.objectID,other.objectID)
	def __nonzero__(self):
		return 1

	def __invokePYRO__(self, *vargs, **kargs):
		if self.adapter is None:
			# delayed adapter binding.
			self.adapter = protocol.getProtocolAdapter(self.URI.protocol)
			self.adapter.bindToURI(self.URI)
			self.adapter.setOneway(self._onewayMethods)
			del self._onewayMethods
		name = self._name[-1]
		del self._name[-1]
		return self.adapter.remoteInvocation(name,
						constants.RIF_Varargs|constants.RIF_Keywords, vargs, kargs)
	# pickling support:
	def __getstate__(self):
		temp = {}
		for key in self.__dict__.keys():
			if key != "adapter":
				temp[key] = self.__dict__[key]
		return temp
	def __setstate__(self, value):
		for key in value.keys():
			self.__dict__[key] = value[key]
		self.adapter = None

	
class DynamicProxyWithAttrs(DynamicProxy):
	def __init__(self, URI):
		# first set the list of 'local' attrs for __setattr__
		self.__dict__["_local_attrs"] = ("_local_attrs","URI", "objectID", "adapter", "_name", "attr_cache","_onewayMethods")
		DynamicProxy.__init__(self, URI)
		self.attr_cache = {}
	def _r_ga(self, attr, value=0):
		if value: self._name.append("_r_ga")
		else: self._name.append("_r_ha")
		return self.__invokePYRO__(attr)
	def findattr(self, attr):
		if self.attr_cache.has_key(attr):
			return self.attr_cache[attr]
		# look it up and cache the value
		self.attr_cache[attr] = self._r_ga(attr)
		return self.attr_cache[attr]
	def __setattr__(self, attr, value):
		if attr in self.__dict__["_local_attrs"]:
			self.__dict__[attr]=value
		else:
			result = self.findattr(attr)
			if result==2: # attribute
				self._name.append("_r_sa")
				return self.__invokePYRO__(attr,value)
			else:
				raise AttributeError('not an attribute')
	def __getattr__(self, attr):
		# allows it to be safely pickeled
		if attr not in ("__getinitargs__", "__hash__","__cmp__","__eq__"):
			result=self.findattr(attr)
			if result==1: # method
				self._name.append(attr)
				return self.__invokePYRO__
			elif result:
				return self._r_ga(attr, 1)
		raise AttributeError()
		

#############################################################################
#
#	Daemon		- server-side Pyro daemon
#
#	Accepts and dispatches incoming Pyro method calls.
#
#############################################################################

class Daemon(protocol.TCPServer, ObjBase):
	def __init__(self,prtcol='PYRO',host='',port=0,norange=0,publishhost=None):
		ObjBase.__init__(self)
		self.setGUID(constants.INTERNAL_DAEMON_GUID)
		self.implementations={constants.INTERNAL_DAEMON_GUID:(self,'__PYRO_Internal_Daemon')}
		if port:
			self.port = port
		else:
			self.port = Pyro.config.PYRO_PORT
		if norange:
			portrange=1
		else:
			portrange=Pyro.config.PYRO_PORT_RANGE
		if not publishhost:
			publishhost=host
		errormsg=''
		for i in range(portrange):
			try:
				protocol.TCPServer.__init__(self, self.port, host, Pyro.config.PYRO_MULTITHREADED)
				self.hostname = publishhost or protocol.getHostname()	
				self.NameServer = None
				self.protocol = prtcol
				self.adapter = protocol.getProtocolAdapter(prtcol)
				self.adapter.setDaemon(self)
				return
			except ProtocolError,msg:
				errormsg=msg
				self.port+=1
		Log.error('Daemon','Couldn\'t start Pyro daemon: ' +str(errormsg))
		raise DaemonError('Couldn\'t start Pyro daemon: ' +str(errormsg))
	
	def __del__(self):
		# server shutting down, unregister all known objects in the NS
		for o in self.implementations.keys():
			self.NameServer.unregister(self.implementations[o][1])

	def __str__(self):
		return '<Pyro Daemon on '+self.hostname+':'+str(self.port)+'>'

	def useNameServer(self,NS):
		self.NameServer=NS

	def getProxyForObj(self, obj):
		return DynamicProxy( PyroURI(self.hostname,
				obj.GUID(), prtcol=self.protocol, port=self.port) )
	def getAttrProxyForObj(self, obj):
		return DynamicProxyWithAttrs( PyroURI(self.hostname,
				obj.GUID(), prtcol=self.protocol, port=self.port) )

	def connectPersistent(self, object, name=None):
		# when a persistent entry is found in the NS, that URI is
		# used instead of the supplied one, if the address matches.
		if name and self.NameServer:
			try:
				newURI = PyroURI(self.hostname, object.GUID(),
					  prtcol=self.protocol, port=self.port)
				URI=self.NameServer.resolve(name)
				if (URI.protocol,URI.address,URI.port)==(newURI.protocol,newURI.address,newURI.port):
					# reuse the previous object ID
					object.setGUID(URI.objectID)
					# enter the (object,name) in the known impl. dictionary
					self.implementations[object.GUID()]=(object,name)
					object.setDaemon(self)
					return URI
				else:
					# name exists, but address etc. is wrong. Remove it.
					# then continue so it wil be re-registered.
					try: self.NameServer.unregister(name)
					except NamingError: pass
			except NamingError:
				pass
		# Register normally.		
		return self.connect(object, name)

	def connect(self, object, name=None):
		URI = PyroURI(self.hostname, object.GUID(),
					  prtcol=self.protocol, port=self.port)
		# if not transient, register the object with the NS
		if name:
			if self.NameServer:
				self.NameServer.register(name, URI)
			else:
				Log.warn('Daemon','connecting object without name server specified:',name)
		# enter the (object,name) in the known implementations dictionary
		self.implementations[object.GUID()]=(object,name)
		object.setDaemon(self)
		return URI

	def disconnect(self,object):
		try:
			if self.implementations[object.GUID()][1]:
				# only unregister with NS if it had a name (was not transient)
				self.NameServer.unregister(self.implementations[object.GUID()][1])
			del self.implementations[object.GUID()]
		finally:
			object.setDaemon(None)

	def handleError(self,conn):			# overridden from TCPServer
		try:
			(exc_type, exc_value, exc_trb) = sys.exc_info()
			if exc_type==ProtocolError:
				# Problem with the communication protocol, shut down the connection
				# XXX is this wat we want???
				Log.error('Daemon','protocol error occured:',exc_value)
				Log.error('Daemon','Due to network error: shutting down connection with',conn.addr)
				self.removeConnection(conn)
			else:
				# This code used to distinguish between 'accidents' and
				# 'on purpose' exceptions. No longer. Every exception
				# is treated the same.
				exclist = traceback.format_exception(exc_type, exc_value, exc_trb)[3:] # strip off Pyro internal traceback
				out =''.join(exclist[-5:]) # print only partial traceback
				Log.warn('Daemon', 'Exception during processing of request from',
					conn.addr,' type',exc_type,
					'\n--- partial traceback of this exception follows:\n',
					out,'\n--- end of traceback')
				self.adapter.returnException(conn.sock,exc_value,0,exclist) # don't shutdown socket
		finally:		
			# clean up circular references to traceback info
			del exc_type, exc_value, exc_trb
	
	def handleRequest(self,conn):
		try:
			self.adapter.handleInvocation(conn)
		except ConnectionClosedError:
			# other side has closed connection
			self.removeConnection(conn)

	def sendAccept(self,sock):
		self.adapter.sendAccept(sock)
	def sendDeny(self,sock,reasonCode):
		self.adapter.sendDeny(sock,reasonCode)

	def ResolvePYROLOC(self, name):
		# this gets called from the protocol adapter when
		# it wants the daemon to resolve a local object name (PYROLOC: protocol)
		Log.msg('Daemon','resolving PYROLOC name: ',name)
		for o in self.implementations.keys():
			if self.implementations[o][1]==name:
				return o
		raise KeyError('no object found by this name')		


#############################################################################
#
#	Client/Server Init code
#
#############################################################################

def _initGeneric_pre():
	if Pyro.config.PYRO_TRACELEVEL == 0: return
	try:
		out='\n'+'-'*60+' NEW SESSION\n'+time.strftime('%Y-%m-%d %H:%M:%S',time.localtime())+ \
			'   Pyro Initializing, version '+constants.VERSION+'\n'
		Log.raw(out)
	except IOError,e:
		sys.stderr.write('PYRO: Can\'t write the tracefile '+Pyro.config.PYRO_LOGFILE.name+'\n'+e)

def _initGeneric_post():
	if Pyro.config.PYRO_TRACELEVEL == 0: return
	try:
		out='Configuration settings are as follows:\n'
		for item in dir(Pyro.config):
			if item[0:4]=='PYRO':
				out+=item+' = '+str(Pyro.config.__dict__[item])+'\n'
		out+='Init done.\n'+'-'*70+'\n'
		Log.raw(out)
	except IOError:
		pass

def initClient(banner=1):
	global _init_client_done
	if _init_client_done: return
	_initGeneric_pre()
	if Pyro.config.PYRO_TRACELEVEL >0: Log.raw('This is initClient.\n')
	_initGeneric_post()
	if banner:
		print 'Pyro Client Initialized. Using Pyro V'+constants.VERSION
	_init_client_done=1
	
def initServer(banner=1):
	global _init_server_done
	if _init_server_done: return
	_initGeneric_pre()
	if Pyro.config.PYRO_TRACELEVEL >0: Log.raw('This is initServer.\n')
	_initGeneric_post()
	if banner:
		print 'Pyro Server Initialized. Using Pyro V'+constants.VERSION
	_init_server_done=1

