#############################################################################
#
#	protocol.py - Pyro Protocol Adapters
#
#	This is part of "Pyro" - Python Remote Objects
#	which is (c) Irmen de Jong - irmen@users.sourceforge.net
#
#############################################################################

import socket, struct, os, time, sys
import Pyro
import util, constants

if os.name!='java':		# Jython has no select module
	import select

from errors import *
from errors import _InternalNoModuleError
pickle = util.pickle
Log = util.Log

if util.supports_multithreading():
	from threading import Thread
	_has_threading = 1
else:
	_has_threading = 0

if util.supports_compression():
	import zlib
	_has_compression = 1
else:
	_has_compression = 0

#------ Get the hostname (possibly of other machines) (returns None on error)
def getHostname(ip=None):
	try:
		if ip:
			(hn,alias,ips) = socket.gethostbyaddr(ip)
			return hn
		else:
			return socket.gethostname()
	except socket.error:
		return None

#------ Get IP address (return None on error)
def getIPAddress(host=None):
	try:
		return socket.gethostbyname(host or getHostname())
	except socket.error:
		return None


#------ Socket helper functions for sending and receiving data correctly.

# Receive a precise number of bytes from a socket. Raises the
# ConnectionClosedError if  that number of bytes was not available.
# (the connection has probably been closed then).
# Never will this function return an empty message (if size>0).
# We need this because 'recv' isn't guaranteed to return all desired
# bytes in one call, for instance, when network load is high.
# Use a list of all chunks and join at the end: faster!
def sock_recvmsg(sock, size):
	msglen=0
	msglist=[]
	try:
		while msglen<size:
			chunk=sock.recv(size-msglen)
			if not chunk:
				raise ConnectionClosedError('connection lost')
			msglist.append(chunk)	
			msglen+=len(chunk)
		return ''.join(msglist)
	except socket.error:
		raise ConnectionClosedError('connection lost')

# Send a message over a socket. Raises ConnectionClosedError if the msg
# couldn't be sent (the connection has probably been lost then).
# We need this because 'send' isn't guaranteed to send all desired
# bytes in one call, for instance, when network load is high.

#def sock_sendmsg(sock, msg):
#	if Pyro.config.PYRO_LARGESOCKMSGBUG:
#		origmsg = msg
#		big = 1024 * 1024
#		bigchunks,leftover = divmod(len(origmsg),big)
#		if leftover:
#			bigchunks +=1
#		for chunk in range(bigchunks):
#			start = chunk * big
#			end = start + big
#			_do_sock_sendmsg(sock, origmsg[start:end])
#	else:
#		_do_sock_sendmsg(sock, msg)

def sock_sendmsg(sock,msg):
	size=len(msg)
	total=0
	try:
		sent=sock.send(msg)
		total+=sent
		if sent==0:
			raise ConnectionClosedError('connection lost')
		while total<size:
			sent=sock.send(msg[total:])
			if sent==0:
				raise ConnectionClosedError('connection lost')
			total+=sent
	except socket.error:
		raise ConnectionClosedError('connection lost')

# set socket option to try to re-use a server port if possible
def set_reuse_addr(sock):
	if os.name!='nt':
		try:
			sock.setsockopt ( socket.SOL_SOCKET, socket.SO_REUSEADDR,
				sock.getsockopt (socket.SOL_SOCKET, socket.SO_REUSEADDR) | 1)
		except:
			pass


#------ PYRO: adapter (default Pyro wire protocol)
#------ This adapter is for protocol version 2.0 ONLY
# Future adapters could be downwards compatible and more flexible.
# Notice that when the header and body format changes, we have
# to increment the protocol VERSION. A revision change is only when
# something inside the body is different but the message as a whole
# can still be received.

PFLG_COMPRESSED = 0x01		# protocol flag: compressed body
PFLG_CHECKSUM = 0x02		# protocol flag: checksum body

class PYROAdapter:
	deniedReasons={ 0:'unspecified reason',1:'server too busy', 2:'host blocked', 3:'security reasons'}
	def __init__(self):
		self.daemon = None			# the Pyro daemon which uses us
		self.headerFmt = '!4sBBHlBl'	# version 3 header
		self.headerID = 'PYRO'
		self.acceptMSG= 'GRANTED'	# must be same length as denyMSG
		self.denyMSG=   'DENIED'	# must be same length as acceptMSG
		                            # (note that a 1-character code is appended)
		self.headerSize = struct.calcsize(self.headerFmt)
		self.version, self.revision = (3,0)	# version 3.0 protocol
		self.onewayMethods=[]		# methods that should be called one-way
	def setDaemon(self, daemon):
		self.daemon = daemon
		Log.msg('PYROAdapter','adapter daemon set to',str(daemon))
	def sendAccept(self, sock):		# called by TCPServer
		sock_sendmsg(sock, self.acceptMSG)
	def sendDeny(self, sock, reasonCode=0):		# called by TCPServer
		sock_sendmsg(sock, self.denyMSG+str(reasonCode)[0])
	def bindToURI(self,URI):
		if URI.protocol not in ('PYRO', 'PYROLOC'):
			Log.error('PYROAdapter','incompatible protocol in URI:',URI.protocol)
			raise ProtocolError('incompatible protocol in URI')
		try:
			self.URI=URI
			sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
			sock.connect((URI.address, URI.port))
			msg = sock_recvmsg(sock, len(self.acceptMSG))
			if msg==self.acceptMSG:
				self.sock=sock
				Log.msg('PYROAdapter','connected to',str(URI))
				if URI.protocol=='PYROLOC':
					objectName=URI.objectID
					Log.msg('PYROAdapter','resolving PYROLOC name: ',objectName)
					# call the special Resolve method on the daemon itself:
					self.URI.objectID=constants.INTERNAL_DAEMON_GUID
					result=self.remoteInvocation('ResolvePYROLOC',0,objectName)
					# found it, switching to regular pyro protocol
					self.URI.objectID=result
					self.URI.protocol='PYRO'
			elif msg[:len(self.denyMSG)]==self.denyMSG:
				try:
					raise ConnectionDeniedError(self.deniedReasons[int(msg[-1])])
				except (KeyError,ValueError):
					raise ConnectionDeniedError('invalid response')
		except socket.error:
			Log.msg('PYROAdapter','connection failed to URI',str(URI))
			raise ProtocolError('connection failed')

	def rebindURI(self, tries=sys.maxint, wait=1):
		t=0
		while t<tries:
			try:
				self.bindToURI(self.URI)
				return
			except ProtocolError:
				t+=1
				if t<tries:
					time.sleep(wait)
		raise ConnectionClosedError('connection lost')
		
	def createMsg(self, body):
		pflgs=0
		if _has_compression and Pyro.config.PYRO_COMPRESSION:
			before=len(body)
			bz=zlib.compress(body) # default compression level
			if len(bz)<before:
				pflgs|=PFLG_COMPRESSED
				body=bz
		crc=0
		if Pyro.config.PYRO_CHECKSUM and _has_compression:
			crc=zlib.adler32(body)
			pflgs|=PFLG_CHECKSUM
		return struct.pack(self.headerFmt, self.headerID,
					self.version, self.revision,
					self.headerSize, len(body), pflgs, crc) + body

	def setOneway(self, methods):
		self.onewayMethods.extend(methods)

	def remoteInvocation(self, method, flags, *args):
		if not self.__dict__.has_key('sock'):
			Log.warn('PYROAdapter','no socket, trying to bind again')
			self.bindToURI(self.URI)	
		if method in self.onewayMethods:
			flags |= constants.RIF_Oneway
		body=pickle.dumps((self.URI.objectID,method,flags,args),Pyro.config.PYRO_BINARY_PICKLE)
		sock_sendmsg(self.sock, self.createMsg(body))
		if flags & constants.RIF_Oneway:
			return None		# no answer required, return immediately
		ver,rev,answer,pflags = self.receiveMsg(self.sock)
		answer = pickle.loads(answer)
		if isinstance(answer,PyroExceptionCapsule):
			if isinstance(answer.excObj,_InternalNoModuleError):
				# server couldn't load module, supply it
				import new
				try:
					importmodule=new.module('-agent-import-')
					mname=answer.excObj.modulename
					try:
						exec 'import '+mname in importmodule.__dict__
					except ImportError,x:
						Log.error('PYROAdapter','Server wanted a non-existing module:',mname)
						raise PyroError('Server wanted a non-existing module',mname)
					m=eval('importmodule.'+mname)
					m=open(m.__file__,'rb').read()
					self.remoteInvocation('remote_supply_code',0,mname, m, self.sock.getsockname())
					# retry the method invocation
					return apply(self.remoteInvocation,(method, flags)+args)
				finally:
					del importmodule
			else:
				# we have an encapsulated exception, raise it again.
				answer.raiseEx()
		return answer

	# receives a socket message, returns: (protocolver, protocolrev, message, protocolflags)
	def receiveMsg(self,sock):
		msg=sock_recvmsg(sock, self.headerSize)
		(id, ver, rev, hsiz, bsiz, pflags, crc) = struct.unpack(self.headerFmt,msg)
		if ver!=self.version:
			Log.error('PYROAdapter','incompatible protocol version')
			# try to report error to client, but most likely the connection will terminate:
			self.returnException(sock, ProtocolError('incompatible protocol version'))
			return ver,rev,None,pflags
		if id!=self.headerID or hsiz!=self.headerSize:
			print 'INVALID HEADER: ',sock.getsockname(),(id, ver, rev, hsiz, bsiz,pflags) # XXX
			# try to report error to client, but most likely the connection will terminate:
			Log.error('PYROAdapter','invalid header')
			self.returnException(sock, ProtocolError('invalid header'))
			return ver,rev,None,pflags
		body=sock_recvmsg(sock, bsiz)
		if pflags&PFLG_CHECKSUM:
			if _has_compression:
				if crc!=zlib.adler32(body):
					Log.error('PYROAdapter','checksum error in body')
					self.returnException(sock, ProtocolError('checksum error'))
					return ver,rev,None,pflags
			else:
				raise ProtocolError('cannot perform checksum')
		if pflags&PFLG_COMPRESSED:
			if _has_compression:
				body=zlib.decompress(body)
			else:
				# XXX We received a compressed message but cannot decompress.
				# Is this really a server error? We now throw an exception on the server...
				raise ProtocolError('compression not supported')
		return ver,rev,body,pflags

	def handleInvocation(self,conn):
		ver,rev,body,pflags = self.receiveMsg(conn.sock)
		if not body:
			# something went wrong even before receiving the full message body
			return
		if ver!=self.version or rev!=self.revision:
			Log.error('PYROAdapter','incompatible protocol version')
			self.returnException(conn.sock, ProtocolError('incompatible protocol version'))
			return 

		# Unpickle the request, which is a tuple:
		#  (object ID, method name, flags, (arg1,arg2,...))
		try:
			req=pickle.loads(body)
		except ImportError,x:
			if Pyro.config.PYRO_MOBILE_CODE:
				# return a special exception that will be processed by client;
				# it will call the internal 'remote_supply_code' member
				self.returnException(conn.sock, _InternalNoModuleError(x.args[0][16:]),0) # don't shutdown!
			else:
				Log.error('PYROAdapter','code problem with incoming object: '+str(x))
				self.returnException(conn.sock, apply(NoModuleError,x.args))
			return
		try:
			# find the object in the implementation database of our daemon
			o = self.daemon.implementations[req[0]]
		except KeyError,x:
			Log.warn('PYROAdapter','Invocation to unknown object ignored:',x)
			self.returnException(conn.sock, ProtocolError('unknown object ID'))
			return
		else:
			# call the method on this object (dispatch in a thread if desired)
			if self.daemon.threaded:
				Thread(target=self._handleInvoc_thread, args=(o[0],req,conn)).start()
			else:
				self._handleInvoc_thread(o[0],req,conn)
	def _handleInvoc_thread(self, proxy, req, conn):
		try:
			flags=req[2]
			res = proxy.Pyro_dyncall(req[1],flags,req[3])	# (method,flags,args)
			if flags&constants.RIF_Oneway:
				return		# no result, return immediately
			# reply the result to the caller
			body=pickle.dumps(res,Pyro.config.PYRO_BINARY_PICKLE)
			sock_sendmsg(conn.sock, self.createMsg(body))
		except:
			self.daemon.handleError(conn)

	def returnException(self, sock, exc, shutdown=1, args=None):
		# return an encapsulated exception to the client
		try:
			body=pickle.dumps(PyroExceptionCapsule(exc,args),Pyro.config.PYRO_BINARY_PICKLE)
		except Exception,x:
			# hmm, pickling the exception failed... pickle the string instead
			body=pickle.dumps(PyroExceptionCapsule(PyroError(str(x)),args),Pyro.config.PYRO_BINARY_PICKLE)
		sock_sendmsg(sock, self.createMsg(body))
		if shutdown:
			sock.shutdown(2) # no further send/receives
	
		
def getProtocolAdapter(protocol):
	if protocol in ('PYRO', 'PYROLOC'):
		return PYROAdapter()
	else:
		Log.error('getProtocolAdapter','unsupported protocol:',protocol)
		raise ProtocolError('unsupported protocol')


#-------- TCPConnection object for TCPServer class
class TCPConnection:
	def __init__(self, sock, addr):
		self.sock = sock
		self.addr = addr
	def fileno(self):
		return self.sock.fileno()
	def close(self):
		self.sock.close()

#-------- new connection validators:
#-------- limitingConnValidator checks max number of connections
class limitingConnValidator:
	def __init__(self, tcpserver):
		self.tcpserver = tcpserver
	def accept(self,connection):
		if len(self.tcpserver.connections)>=Pyro.config.PYRO_MAXCONNECTIONS:
			Log.msg('limitingConnValidator','Too many open connections, closing',connection.addr,'#conns=',len(self.tcpserver.connections))
			return (0, 1) # reason 1 = server full
		return (1,0)

#-------- TCPServer base class
class TCPServer:
	def __init__(self, port, host='', threaded=_has_threading):
		try:
			if os.name=='java':
				raise NotImplementedError('Pyro server not yet supported on Jython') # XXX
			self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
			set_reuse_addr(self.sock)
			self.sock.bind((host,port))
			self.sock.listen(5)
			self.connections = []
			self.threaded = threaded
			self.newConnValidator=limitingConnValidator(self)
		except socket.error,msg:
			raise ProtocolError(msg)
			
	def __del__(self):
		if len(self.connections)>0:
			Log.warn('TCPServer','Shutting down but there are still',len(self.connections),'active connections')
		
	# this is the preferred way of dealing with the request loop.
	def requestLoop(self, condition=lambda:1, timeout=3, others=[], callback=None):
		while condition():
			self.handleRequests(timeout,others,callback)

	def setNewConnectionValidator(self,validator):
		self.newConnValidator=validator
	def newConnection(self,sock):
		csock, addr = sock.accept()
		conn=TCPConnection(csock,addr)
		(ok,reasonCode) = self.newConnValidator.accept(conn)
		if ok:
			self.sendAccept(csock)
			self.connections.append(conn)
			Log.msg('TCPServer','new connection from',addr,'#conns=',len(self.connections))
		else:
			# log entry has already been written by newConnValidator
			self.sendDeny(csock,reasonCode)
			conn.close()

	def handleRequests(self, timeout=None, others=[], callback=None):
		handledRequest=0	
		activeCnt=1
		while activeCnt and not handledRequest:
			socklist = self.connections+[self.sock]+others
			if timeout==None:
				ins,outs,exs = select.select(socklist,[],[])
			else:
				ins,outs,exs = select.select(socklist,[],[],timeout)
			activeCnt=len(ins)
			if self.sock in ins:
				self.newConnection(self.sock)
				ins.remove(self.sock)
			for c in ins[0:]:
				if isinstance(c,TCPConnection):
					try:
						self.handleRequest(c)
					except:
						self.handleError(c)
					ins.remove(c)
					handledRequest=1
			if ins and callback:
				# the 'others' must have fired...
				callback(ins)

	# to be called if a dropped connection is detected:
	def removeConnection(self, conn):
		if conn in self.connections:
			self.connections.remove(conn)
			Log.msg('TCPServer','removed connection with',conn.addr,' #conns=',len(self.connections))

	def sendAccept(self,csock):
		raise NotImplementedError,'must be overridden'
	def sendDeny(self,csock,reasonCode):
		raise NotImplementedError,'must be overridden'
	def handleError(self,conn):
		raise NotImplementedError,'must be overridden'
	def handleRequest(self,conn):
		raise NotImplementedError,'must be overridden'

