#############################################################################
#
#	util.py - Pyro Utilities
#
#	This is part of "Pyro" - Python Remote Objects
#	which is (c) Irmen de Jong - irmen@users.sourceforge.net
#
#############################################################################

# Fall back on pickle if cPickle isn't available
try: 
	import cPickle; pickle = cPickle
except ImportError:
	import pickle

import os,random,time,string,sys,time,Pyro


# Log classes for (M)essages, (W)arnings, (E)rrors.


class LoggerBase:
	# Logger base class. Subclasses must override _logfile and  _checkTraceLevel.
	def msg(self,source,*args):
		if self._checkTraceLevel(3): self._trace('NOTE',source, args)
	def warn(self,source,*args):
		if self._checkTraceLevel(2): self._trace('WARN',source, args)
	def error(self,source,*args):
		if self._checkTraceLevel(1): self._trace('ERR!',source, args)
	def raw(self,str):
		try:
			open(self._logfile(),'a').write(str)
		except:
			pass
	def _trace(self,typ,source, arglist):
		if not arglist:
			arglist=['<no message>']
		try:
			tf=open(self._logfile(),'a')
			tf.write(time.strftime('%Y-%m-%d %H:%M:%S',time.localtime(time.time()))+
			  ' ** '+typ+' ** '+str(source)+' ** '+reduce(lambda x,y: str(x)+' '+str(y),arglist)+'\n')
		except:
			pass
	def _logfile(self):
		raise NotImplementedError,'must override'
	def _checkTraceLevel(self,level):
		raise NotImplementedError,'must override'

class SystemLogger(LoggerBase):
	def _checkTraceLevel(self, level):
		return Pyro.config.PYRO_TRACELEVEL >= level
	def _logfile(self):
		file=Pyro.config.PYRO_LOGFILE
		if not os.path.isabs(file):
			Pyro.config.PYRO_LOGFILE=os.path.join(Pyro.config.PYRO_STORAGE, file)
		return Pyro.config.PYRO_LOGFILE
		
class UserLogger(LoggerBase):
	def _checkTraceLevel(self, level):
		return Pyro.config.PYRO_USER_TRACELEVEL >= level
	def _logfile(self):
		file=Pyro.config.PYRO_USER_LOGFILE
		if not os.path.isabs(file):
			Pyro.config.PYRO_USER_LOGFILE=os.path.join(Pyro.config.PYRO_STORAGE, file)
		return Pyro.config.PYRO_USER_LOGFILE


# The logger object 'Log'.
Log = SystemLogger()


# Caching directory lister, outputs (filelist,dirlist) tuple
# Based upon dircache.py
_listdir_cache = {}

def listdir(path):
	try:
		cached_mtime, files, directories = _listdir_cache[path]
		del _listdir_cache[path]
	except KeyError:
		cached_mtime, files, directories = -1, [], []
	try:
		mtime = os.stat(path)[8]
	except os.error:
		return () 
	if mtime <> cached_mtime:
		try:
			list = os.listdir(path)
		except os.error:
			return ()
		files=[]
		directories=[]
		for e in list:
			if os.path.isdir(os.path.join(path,e)):
				directories.append(e)
			else:
				files.append(e)
	_listdir_cache[path] = mtime, files, directories
	return files,directories


# Fairly simple argument options parser. Like getopt(3).
class ArgParser:
	def __init__(self):
		pass
	def parse(self, args, optionlist):
		# arglist is a string such as "ab:c" which means
		# we search for 3 options (-a, -b, -c) of which -b has an argument.
		self.options={}		# public, the option->value dictionary
		self.args=[]		# public, the rest of the arguments
		self.ignored=[]		# public, ignored options
		optionlist+=' '		# add sentinel
		if type(args)==type(''):
			args=string.split(args)
		while args:
			arg=args[0]
			del args[0]
			if arg[0]=='-':
				if len(arg)>=2:   # arg is an option. Check our list
					idx = string.find(optionlist,arg[1])
					if idx>=0:
						if optionlist[idx+1]==':':   # option requires argument.
							if len(arg)>=3:   # argument is appended. Use this.
								self.options[arg[1]]=arg[2:]
								continue
							# fetch argument from next string
							if len(args)>=1:
								self.options[arg[1]]=args[0]
								del args[0]
								continue
							else:   # missing arg, substitute None
								self.options[arg[1]]=None
						else:   # option requires no argument, use None
							self.options[arg[1]]=None
					else:   # didn't find this option, skip it
						self.ignored.append(arg[1])
				else:   # arg is a single '-'. Stop parsing.
					for a in args:
						self.args.append(a)
					args=None
			else:   # arg is no option, add it to the residu list and continue
				self.args.append(arg)
	def getOpt(self, option, default=Exception()):
		try:
			return self.options[option]
		except KeyError:
			if not isinstance(default,Exception):
				return default
			raise KeyError('no such option')
	def printIgnored(self):
		if self.ignored:
			print 'Ignored options:',
			for o in self.ignored:
				print '-'+o,
			print


def int2binstr(i,l):
	s=''
	for a in range(l):
		s=chr(i&0xFF)+s
		i>>=8
	return s

def binstr2int(s):
	i=0
	for c in s:
		i=(i<<8) + ord(c)
	return i
	
	
def getBinaryGUID():
	# Generate binary GUID.
	# The GUID is constructed as follows;
	# AAAAAAAA-AAAABBBB-BBBBBBBB-BBCCCCCC  (a 128-bit number in hex)
	# where A=network address, B=timestamp, C=random. 
	# The 128 bit number is returned as a string of 16 8-bits characters.
	# For A: should use the machine's MAC ethernet address, but there is no
	# portable way to get it... use the IP address + 2 bytes process id.

	from Pyro.protocol import getIPAddress
	
	ip=getIPAddress()
	if ip:
		ip=map(int, string.split(ip,'.'))
		ip=reduce(lambda x,y: (x<<8)+y, ip, 0L)
	else:
		# can't get IP address... use another value, like our Python id() and PID
		Log.warn('getGUID','Can\'t get IP address')
		try:
			ip=os.getpid()
		except:
			ip=0
		ip += id(getBinaryGUID)

	if os.name!='java':
		extra=os.getpid()
	else:
		extra=id(os)		# XXX preliminary Jython compat. (no getpid)

	networkAddr = (long(ip) << 16 ) | extra
	t1=time.time()*100
	t2=int((t1*time.clock())%sys.maxint) & 0xffffff
	t1=int(t1%sys.maxint) 
	timestamp = (long(t1) << 24) | t2
	r2=(random.randint(0,sys.maxint/2)>>4) & 0xffff
	r3=(random.randint(0,sys.maxint/2)>>5) & 0xff
	rndm = (r2<<8) | r3
	return int2binstr(networkAddr,6)+int2binstr(timestamp,7)+int2binstr(rndm,3)

# bin2ascGUID converts a binary guid from getBinaryGUID to a readable string:
# AAAAAAAA-AAAABBBB-BBBBBBBB-BBCCCCCC
def bin2ascGUID(guid):
	i1 = binstr2int(guid[0:4])
	i2 = binstr2int(guid[4:8])
	i3 = binstr2int(guid[8:12])
	i4 = binstr2int(guid[12:16])
	return '%08lx-%08lx-%08lx-%08lx' % (i1,i2,i3,i4)

# asc2binGUID converts a string guid to a binary one (16-byte string):
def asc2binGUID(guid):
	try:
		return int2binstr(string.atol(guid[:8],16),4) + \
			int2binstr(string.atol(guid[9:17],16),4) + \
			int2binstr(string.atol(guid[18:26],16),4) + \
			int2binstr(string.atol(guid[27:],16),4) 
	except SyntaxError:
		raise ValueError('invalid GUID format')

		
def genguid_scripthelper():
	if __name__!='__main__':
		del sys.argv[0]	# remove '-c' arg
	p=ArgParser()
	p.parse(sys.argv,'')
	if p.args or p.ignored:
		print 'usage: genguid'
		raise SystemExit
	g=getBinaryGUID()
	print bin2ascGUID(g)


def supports_multithreading():
	try:
		import threading
		return 1
	except:
		return 0
	
def supports_compression():
	try:
		import zlib
		return 1
	except:
		return 0


# Pyro traceback printing
def getPyroTraceback(exc_obj):
	import traceback, constants, sys
	try:
		exc_type, exc_value, exc_trb=sys.exc_info()
		remote_tb=getattr(exc_obj,constants.TRACEBACK_ATTRIBUTE,None)
		l_tb_list=traceback.format_exception(exc_type, exc_value, exc_trb)
		local_tb=traceback.format_exception(exc_type, exc_value, exc_trb)
		if remote_tb:
			return [ '---- remote Pyro traceback ----\n' ] + \
				remote_tb + [ '---- local traceback ----\n' ] +  local_tb
		else:
			# hmm. no remote tb info, return just the local tb.
			return local_tb
	finally:
		# clean up cycle to traceback
		del exc_type, exc_value, exc_trb


